"""
Udisks wrapper utilities.

These act as a convenience abstraction layer on the udisks dbus service.
Requires Udisks 1.0.5 as described here:

    http://udisks.freedesktop.org/docs/1.0.5/

Note that (as this completely wraps the udisks dbus API) replacing this
module will let you support Udisks2 or maybe even other services.

"""
__all__ = ['Device', 'get_all', 'get_all_handleable', 'get_device']

import logging
import os
import dbus

from udiskie.common import Properties as DbusProperties


UDISKS_INTERFACE = 'org.freedesktop.UDisks'
UDISKS_DEVICE_INTERFACE = 'org.freedesktop.UDisks.Device'

UDISKS_OBJECT = 'org.freedesktop.UDisks'
UDISKS_OBJECT_PATH = '/org/freedesktop/UDisks'


class Device:
    """
    Wrapper class for org.freedesktop.UDisks.Device proxy objects.
    """
    def __init__(self, bus, device_path):
        self.log = logging.getLogger('udiskie.udisks.Device')
        self.bus = bus
        self.device_path = device_path
        device_object = self.bus.get_object(UDISKS_OBJECT, device_path)
        self.property = DbusProperties(device_object, UDISKS_DEVICE_INTERFACE)
        self.method = dbus.Interface(device_object, UDISKS_DEVICE_INTERFACE)

    def __str__(self):
        return self.device_path

    # properties
    @property
    def partition_slave(self):
        """Get the partition slave."""
        return self.property.PartitionSlave

    @property
    def is_partition_table(self):
        """Check if the device is a partition table."""
        return self.property.DeviceIsPartitionTable

    @property
    def is_systeminternal(self):
        """Check if the device is internal."""
        return self.property.DeviceIsSystemInternal

    @property
    def is_handleable(self):
        """
        Should this device be handled by udiskie?

        Currently this just means that the device is removable and holds a
        filesystem or the device is a LUKS encrypted volume.

        """
        return (self.is_filesystem or self.is_crypto) and not self.is_systeminternal

    @property
    def is_mounted(self):
        """Check if the device is mounted."""
        return self.property.DeviceIsMounted

    @property
    def is_unlocked(self):
        """Check if device is already unlocked."""
        return self.is_luks and self.luks_cleartext_holder

    @property
    def mount_paths(self):
        raw_paths = self.property.DeviceMountPaths
        return [os.path.normpath(path) for path in raw_paths]

    @property
    def device_file(self):
        return os.path.normpath(self.property.DeviceFile)

    @property
    def is_filesystem(self):
        return self.id_usage == 'filesystem'

    @property
    def is_crypto(self):
        return self.id_usage == 'crypto'

    @property
    def is_luks(self):
        return self.property.DeviceIsLuks

    @property
    def is_luks_cleartext(self):
        """Check whether this is a luks cleartext device."""
        return self.property.DeviceIsLuksCleartext

    @property
    def luks_cleartext_slave(self):
        """Get luks crypto device."""
        return self.property.LuksCleartextSlave

    @property
    def luks_cleartext_holder(self):
        """Get unlocked luks cleartext device."""
        return self.property.LuksHolder

    @property
    def is_luks_cleartext_slave(self):
        """Check whether the luks device is currently in use."""
        if not self.is_luks:
            return False
        for device in get_all(self.bus):
            if (not device.is_filesystem or device.is_mounted) and (
                    device.is_luks_cleartext and
                    device.luks_cleartext_slave == self.device_path):
                return True
        return False

    @property
    def has_media(self):
        return self.property.DeviceIsMediaAvailable

    @property
    def id_type(self):
        return self.property.IdType

    @property
    def id_usage(self):
        return self.property.IdUsage

    @property
    def id_uuid(self):
        return self.property.IdUuid

    # methods
    def mount(self, filesystem, options):
        self.method.FilesystemMount(filesystem, options)

    def unmount(self):
        self.method.FilesystemUnmount([])

    def lock(self, options):
        """Lock Luks device."""
        return self.method.LuksLock(options)

    def unlock(self, password, options):
        """Unlock Luks device."""
        return self.method.LuksUnlock(password, options)


def get_all(bus):
    """Enumerate all device objects currently known to udisks."""
    udisks = bus.get_object(UDISKS_OBJECT, UDISKS_OBJECT_PATH)
    for path in udisks.EnumerateDevices(dbus_interface=UDISKS_INTERFACE):
        yield Device(bus, path)

def get_all_handleable(bus):
    """Enumerate all handleable devices currently known to udisks."""
    for device in get_all(bus):
        if device.is_handleable:
            yield device

def get_device(bus, path):
    """Get a device proxy by device name or any mount path of the device."""
    logger = logging.getLogger('udiskie.udisks.get_device')
    for device in get_all(bus):
        if os.path.samefile(path, device.device_file):
            return device
        for p in device.mount_paths:
            if os.path.samefile(path, p):
                return device
    logger.warn('Device not found: %s' % path)
    return None




"""
Udiskie notification daemon.
"""
__all__ = ['Notify']

import pynotify
import gio

class Notify:
    """
    Notification tool.

    Can be connected to udisks daemon in order to automatically issue
    notifications when system status has changed.

    """
    def __init__(self, name):
        """Initialize notifier."""
        pynotify.init(name)

    # event handlers:
    def device_mounted(self, device):
        try:
            device_file = device.device_file
            mount_path = device.mount_paths[0]
            pynotify.Notification('Device mounted',
                                  '%s mounted on %s' % (device_file, mount_path),
                                  'drive-removable-media').show()
        except gio.Error:
            pass

    def device_unmounted(self, device):
        try:
            device_file = device.device_file
            pynotify.Notification('Device unmounted',
                                  '%s unmounted' % (device_file,),
                                  'drive-removable-media').show()
        except gio.Error:
            pass

    def device_locked(self, device):
        try:
            device_file = device.device_file
            pynotify.Notification('Device locked',
                                  '%s locked' % (device_file,),
                                  'drive-removable-media').show()
        except gio.Error:
            pass

    def device_unlocked(self, device):
        try:
            device_file = device.device_file
            pynotify.Notification('Device unlocked',
                                  '%s unlocked' % (device_file,),
                                  'drive-removable-media').show()
        except gio.Error:
            pass




"""
Common utilities.
"""
__all__ = ['Properties', 'get_udisks', 'default_udisks']
import dbus

DBUS_PROPS_INTERFACE = 'org.freedesktop.DBus.Properties'
default_udisks = 'udiskie.udisks'


def get_udisks():
    import importlib
    return importlib.import_module(default_udisks)

class Properties:
    """
    Dbus property map abstraction.

    Properties of the object can be accessed as attributes.

    """
    def __init__(self, dbus_object, interface):
        """Initialize a proxy object with standard dbus property interface."""
        self.__proxy = dbus.Interface(
                dbus_object,
                dbus_interface=DBUS_PROPS_INTERFACE)
        self.__interface = interface

    def __getattr__(self, property):
        """Retrieve the property via the dbus proxy."""
        return self.__proxy.Get(self.__interface, property)




"""
Udiskie CLI logic.
"""
__all__ = [
    'load_filter',
    'mount_program_options', 'mount',
    'umount_program_options', 'umount']

import warnings
warnings.filterwarnings("ignore", ".*could not open display.*", Warning)
warnings.filterwarnings("ignore", ".*g_object_unref.*", Warning)

import os
import logging
import dbus

import udiskie.match
import udiskie.mount
import udiskie.prompt
import udiskie.notify
import udiskie.automount
import udiskie.daemon
import udiskie.common


CONFIG_PATH = 'udiskie/filters.conf'

def load_filter(filter_file=None):
    """Load mount option filters."""
    try:
        from xdg.BaseDirectory import xdg_config_home
    except ImportError:
        xdg_config_home = os.path.expanduser('~/.config')
    if not filter_file:
        filter_file = os.path.join(xdg_config_home, CONFIG_PATH)
    return udiskie.match.FilterMatcher((filter_file,))


def mount_program_options():
    """
    Return the mount option parser for the mount command.
    """
    import optparse
    parser = optparse.OptionParser()
    parser.add_option('-a', '--all', action='store_true',
                      dest='all', default=False,
                      help='mount all present devices')
    parser.add_option('-v', '--verbose', action='store_const',
                      dest='log_level', default=logging.INFO,
                      const=logging.DEBUG, help='verbose output')
    parser.add_option('-f', '--filters', action='store',
                      dest='filters', default=None,
                      metavar='FILE', help='filter FILE')
    parser.add_option('-s', '--suppress', action='store_true',
                      dest='suppress_notify', default=False,
                      help='suppress popup notifications')
    parser.add_option('-P', '--password-prompt', action='store',
                      dest='password_prompt', default='zenity',
                      metavar='MODULE', help="replace password prompt")
    return parser


def mount(args, allow_daemon=False):
    """
    Execute the mount/daemon command.
    """
    parser = mount_program_options()
    options, posargs = parser.parse_args(args)
    logging.basicConfig(level=options.log_level, format='%(message)s')
    run_daemon = allow_daemon and not options.all and len(posargs) == 0

    # establish connection to system bus
    if run_daemon:
        from dbus.mainloop.glib import DBusGMainLoop
        DBusGMainLoop(set_as_default=True)
    bus = dbus.SystemBus()

    # for now: just use the default udisks
    udisks = udiskie.common.get_udisks()

    # create a mounter
    prompt = udiskie.prompt.password(options.password_prompt)
    filter = load_filter(options.filters)
    mounter = udiskie.mount.Mounter(bus=bus, filter=filter, prompt=prompt, udisks=udisks)

    # run udiskie daemon if needed
    if run_daemon:
        daemon = udiskie.daemon.Daemon(bus, udisks=udisks)
    if run_daemon and not options.suppress_notify:
        notify = udiskie.notify.Notify('udiskie.mount')
        daemon.connect(notify)
    if run_daemon:
        automount = udiskie.automount.AutoMounter(mounter)
        daemon.connect(automount)

    # mount all present devices
    if options.all:
        mounter.mount_all()

    # only mount the desired devices
    elif len(posargs) > 0:
        mounted = []
        for path in posargs:
            device = mounter.mount(path)
            if device:
                mounted.append(device)
        # automatically mount luks holders
        for device in mounted:
            mounter.mount_holder(device)

    # run in daemon mode
    elif run_daemon:
        mounter.mount_all()
        try:
            return daemon.run()
        except KeyboardInterrupt:
            pass

    # print command line options
    else:
        parser.print_usage()
        return 1


def umount_program_options():
    """
    Return the command line option parser for the umount command.
    """
    import optparse
    parser = optparse.OptionParser()
    parser.add_option('-a', '--all', action='store_true',
                      dest='all', default=False,
                      help='all devices')
    parser.add_option('-v', '--verbose', action='store_const',
                      dest='log_level', default=logging.INFO,
                      const=logging.DEBUG, help='verbose output')
    parser.add_option('-s', '--suppress', action='store_true',
                      dest='suppress_notify', default=False,
                      help='suppress popup notifications')
    return parser


def umount(args):
    """
    Execute the umount command.
    """
    parser = umount_program_options()
    (options, posargs) = parser.parse_args(args)
    logging.basicConfig(level=options.log_level, format='%(message)s')
    bus = dbus.SystemBus()

    if len(posargs) == 0 and not options.all:
        parser.print_usage()
        return 1

    # for now: use udisks v1 service
    udisks = udiskie.common.get_udisks()

    if options.all:
        unmounted = udiskie.mount.unmount_all(bus=bus, udisks=udisks)
    else:
        unmounted = []
        for path in posargs:
            device = udiskie.mount.unmount(os.path.normpath(path), bus=bus, udisks=udisks)
            if device:
                unmounted.append(device)

    # automatically lock unused luks slaves of unmounted devices
    for device in unmounted:
        udiskie.mount.lock_slave(device, udisks=udisks)




"""
Udiskie unmount utilities.
"""
__all__ = [
    'unmount_device', 'lock_device', 'remove_device', 'lock_slave',
    'unmount', 'unmount_all']

from udiskie.mount import unmount_device, lock_device, remove_device, lock_slave, unmount, unmount_all




"""
Udiskie automounter daemon.
"""
__all__ = ['AutoMounter']

class AutoMounter:
    """
    Automatically mount newly added media.
    """
    def __init__(self, mounter):
        self.mounter = mounter

    def device_added(self, udevice):
        self.mounter.add_device(udevice)

    def media_added(self, udevice):
        self.mounter.add_device(udevice)




"""
Udiskie mount utilities.
"""
__all__ = [
    'mount_device', 'unmount_device',
    'unlock_device', 'lock_device',
    'add_device', 'remove_device',
    'mount_all', 'unmount_all',
    'mount', 'unmount',
    'mount_holder', 'lock_slave',
    'Mounter']

import sys
import logging
import dbus

from udiskie.common import get_udisks


# mount/unmount
def mount_device(device, filter=None):
    """
    Mount the device if not already mounted.

    Return value indicates whether an action was performed successfully.
    The special value `None` means unknown/unreliable.

    """
    log = logging.getLogger('udiskie.mount.mount_device')
    if not device.is_handleable or not device.is_filesystem:
        log.debug('skipping unhandled device %s' % (device,))
        return False
    if device.is_mounted:
        log.debug('skipping mounted device %s' % (device,))
        return False

    fstype = str(device.id_type)
    options = filter.get_mount_options(device) if filter else []

    S = 'attempting to mount device %s (%s:%s)'
    log.info(S % (device, fstype, options))

    try:
        device.mount(fstype, options)
        log.info('mounted device %s' % (device,))
    except dbus.exceptions.DBusException:
        err = sys.exc_info()[1]
        log.error('failed to mount device %s: %s' % (device, err))
        return None

    mount_paths = ', '.join(device.mount_paths)
    return True

def unmount_device(device):
    """
    Unmount a Device.

    Checks to make sure the device is unmountable and then unmounts.
    Return value indicates whether an action was performed successfully.
    The special value `None` means unknown/unreliable.

    """
    logger = logging.getLogger('udiskie.mount.unmount_device')
    if not device.is_handleable or not device.is_filesystem:
        logger.debug('skipping unhandled device %s' % (device,))
        return False
    if not device.is_mounted:
        logger.debug('skipping unmounted device %s' % (device,))
        return False
    try:
        device.unmount()
        logger.info('unmounted device %s' % (device,))
    except dbus.exceptions.DBusException:
        err = sys.exc_info()[1]
        logger.error('failed to unmount device %s: %s' % (device, err))
        return None
    return True


# unlock/lock (LUKS)
def unlock_device(device, prompt, udisks=None):
    """
    Unlock the device if not already unlocked.

    Return value indicates whether an action was performed successfully.
    The special value `None` means unknown/unreliable.

    """
    log = logging.getLogger('udiskie.mount.unlock_device')
    if not device.is_handleable or not device.is_crypto:
        log.debug('skipping unhandled device %s' % (device,))
        return False
    if device.is_unlocked:
        log.debug('skipping unlocked device %s' % (device,))
        return False

    # prompt user for password
    password = prompt and prompt(
            'Enter password for %s:' % (device,),
            'Unlock encrypted device')
    if password is None:
        return False

    # unlock device
    log.info('attempting to unlock device %s' % (device,))
    try:
        device.unlock(password, [])
        udisks = udisks or get_udisks()
        holder_dev = udisks.Device(device.bus, device.luks_cleartext_holder)
        holder_path = holder_dev.device_file
        log.info('unlocked device %s on %s' % (device, holder_path))
    except dbus.exceptions.DBusException:
        err = sys.exc_info()[1]
        log.error('failed to unlock device %s:\n%s' % (device, err))
        return None
    return True

def lock_device(device):
    """
    Lock device.

    Checks to make sure the device is lockable, then locks.
    Return value indicates whether an action was performed successfully.
    The special value `None` means unknown/unreliable.

    """
    logger = logging.getLogger('udiskie.mount.lock_device')
    if not device.is_handleable or not device.is_crypto:
        logger.debug('skipping unhandled device %s' % (device,))
        return False
    if not device.is_unlocked:
        logger.debug('skipping locked device %s' % (device,))
        return False
    try:
        device.lock([])
        logger.info('locked device %s' % (device,))
    except dbus.exceptions.DBusException:
        err = sys.exc_info()[1]
        logger.error('failed to lock device %s: %s' % (device, err))
        return None
    return True


# add/remove (unlock/lock or mount/unmount)
def add_device(device, filter=None, prompt=None, udisks=None):
    """Mount or unlock the device depending on its type."""
    log = logging.getLogger('udiskie.mount.add_device')
    if not device.is_handleable:
        log.debug('skipping unhandled device %s' % (device,))
        return False
    if device.is_filesystem:
        return mount_device(device, filter)
    elif device.is_crypto:
        return unlock_device(device, prompt, udisks=udisks)

def remove_device(device):
    """Unmount or lock the device depending on device type."""
    logger = logging.getLogger('udiskie.mount.remove_device')
    if not device.is_handleable:
        logger.debug('skipping unhandled device %s' % (device,))
        return False
    if device.is_filesystem:
        return unmount_device(device)
    elif device.is_crypto:
        return lock_device(device)

# mount_all/unmount_all
def mount_all(bus=None, filter=None, prompt=None, udisks=None):
    """Mount handleable devices that are already present."""
    bus = bus or dbus.SystemBus()
    udisks = udisks or get_udisks()
    for device in udisks.get_all_handleable(bus):
        add_device(device, filter, prompt, udisks=udisks)

def unmount_all(bus=None, udisks=None):
    """Unmount all filesystems handleable by udiskie."""
    unmounted = []
    bus = bus or dbus.SystemBus()
    udisks = udisks or get_udisks()
    for device in udisks.get_all_handleable(bus):
        if unmount_device(device):
            unmounted.append(device)
    return unmounted


# mount a holder/lock a slave
def mount_holder(device, filter=None, prompt=None, udisks=None):
    """
    Mount or unlock the holder device of this unlocked LUKS device.

    Will not mount the holder if the device is not unlocked.
    Return value indicates success

    """
    logger = logging.getLogger('udiskie.mount.lock_slave')
    if not device.is_unlocked:
        logger.debug('skipping locked or non-luks device %s' % (device,))
        return False
    udisks = udisks or get_udisks()
    holder_path = device.luks_cleartext_holder
    holder = udisks.Device(device.bus, holder_path)
    return add_device(holder, filter=filter, prompt=prompt, udisks=udisks)

def lock_slave(device, udisks=None):
    """
    Lock the luks slave of this device.

    Will not lock the slave if it is still used by any mounted file system.
    Return value indicates success.

    """
    logger = logging.getLogger('udiskie.mount.lock_slave')
    if not device.is_luks_cleartext:
        logger.debug('skipping non-luks-cleartext device %s' % (device,))
        return False
    udisks = udisks or get_udisks()
    slave_path = device.luks_cleartext_slave
    slave = udisks.Device(device.bus, slave_path)
    if slave.is_luks_cleartext_slave:
        return False
    return lock_device(slave)


# mount/unmount by path
def mount(path, bus=None, filter=None, prompt=None, udisks=None):
    """
    Mount or unlock a device.

    The device must match the criteria for a filesystem mountable or
    unlockable by udiskie.

    """
    logger = logging.getLogger('udiskie.mount.unmount')
    bus = bus or dbus.SystemBus()
    udisks = udisks or get_udisks()
    device = udisks.get_device(bus, path)
    if device:
        logger.debug('found device owning "%s": "%s"' % (path, device))
        if add_device(device, filter=filter, prompt=prompt, udisks=udisks):
            return device
    return None

def unmount(path, bus=None, udisks=None):
    """
    Unmount or lock a filesystem

    The filesystem must match the criteria for a filesystem mountable by
    udiskie.  path is either the physical device node (e.g. /dev/sdb1) or the
    mount point (e.g. /media/Foo).

    """
    logger = logging.getLogger('udiskie.mount.unmount')
    bus = bus or dbus.SystemBus()
    udisks = udisks or get_udisks()
    device = udisks.get_device(bus, path)
    if device:
        logger.debug('found device owning "%s": "%s"' % (path, device))
        if remove_device(device):
            return device
    return None


# utility class
class Mounter:
    """
    Mount utility.

    Calls the global functions and remembers bus, filter and prompt as well
    as the desired udisks module.

    """
    def __init__(self, bus, filter=None, prompt=None, udisks=None):
        self.bus = bus or dbus.SystemBus()
        self.filter = filter
        self.prompt = prompt
        self.udisks = udisks

    # mount/unmount
    def mount_device(self, device, filter=None):
        return mount_device(device, filter=filter or self.filter)
    def unmount_device(self, device):
        return unmount_device(device)

    # unlock/lock (LUKS)
    def unlock_device(self, device, prompt=None, udisks=None):
        return mount_device(
                device,
                filter=prompt or self.prompt,
                udisks=udisks or self.udisks)
    def lock_device(self, device):
        return lock_device(device)

    # add/remove (unlock/lock or mount/unmount)
    def add_device(self, device, filter=None, prompt=None, udisks=None):
        return add_device(
                device,
                filter=filter or self.filter,
                prompt=prompt or self.prompt,
                udisks=udisks or self.udisks)
    def remove_device(self, device):
        return remove_device(device)

    # mount_all/unmount_all
    def mount_all(self, filter=None, prompt=None, udisks=None):
        return mount_all(
                self.bus,
                filter=filter or self.filter,
                prompt=prompt or self.prompt,
                udisks=udisks or self.udisks)
    def unmount_all(self, udisks=None):
        return unmount_all(self.bus, udisks=udisks or self.udisks)

    # mount/unmount
    def mount(self, path, filter=None, prompt=None):
        return mount(
                path, bus=self.bus,
                filter=filter or self.filter,
                prompt=prompt or self.prompt)
    def unmount(self, path):
        return unmount(path, bus=self.bus)

    # mount_holder/lock_slave
    def mount_holder(self, device, filter=None, prompt=None, udisks=None):
        return mount_holder(
                device,
                filter=filter or self.filter,
                prompt=prompt or self.prompt,
                udisks=udisks or self.udisks)
    def lock_slave(self, device, udisks=None):
        return lock_slave(device, udisks=udisks or self.udisks)




# encoding: utf-8
from __future__ import print_function
from setuptools import setup
import sys

# check availability of runtime dependencies
import argparse
parser = argparse.ArgumentParser()
parser.add_argument('-f', '--force', dest="force",
        action="store_true", default=False)
args, sys.argv = parser.parse_known_args(sys.argv)
if not args.force:
    try:
        import dbus
        import gobject
        import pynotify
    except ImportError:
        err = sys.exc_info()[1]
        print("Missing runtime dependency:", err)
        print("Use --force if you want to continue anyway.")
        sys.exit(1)

# read long_description from README.rst
try:
    f = open('README.rst')
    long_description = f.read()
    f.close()
except:
    long_description = None

setup(
    name='udiskie',
    version='0.5.0-1',
    description='Removable disk automounter for udisks',
    long_description=long_description,
    author='Byron Clark',
    author_email='byron@theclarkfamily.name',
    maintainer='Thomas Gläßle',
    maintainer_email='t_glaessle@gmx.de',
    url='https://github.com/coldfix/udiskie',
    license='MIT',
    packages=[
        'udiskie',
    ],
    scripts=[
        'bin/udiskie',
        'bin/udiskie-umount',
        'bin/udiskie-mount'
    ],
    install_requires=[
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Environment :: X11 Applications',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'License :: OSI Approved :: MIT License',
        'Topic :: Desktop Environment',
        'Topic :: Software Development',
        'Topic :: System :: Filesystems',
        'Topic :: System :: Hardware',
        'Topic :: Utilities',
    ],
)
